@@ -1,4 +1,4 @@
-from typing import Dict, Literal, Tuple, TypeVar, Union
+from typing import ClassVar, Dict, Literal, Tuple, TypeVar, Union
 
 import casadi as cs
 import numpy as np
@@ -13,7 +13,7 @@ T = TypeVar("T", cs.SX, cs.MX)
 class HasConstraints(HasVariables[T]):
     """Class for creating and storing symbolic constraints for an NLP problem."""
 
-    _DUAL_VARIABLES_ORDER = ("g", "h", "h_lbx", "h_ubx")
+    dual_variables_order: ClassVar[Tuple[str, ...]] = ("g", "h", "h_lbx", "h_ubx")
     """This tuple dictates the order for operations related to dual variables."""
 
     def __init__(
@@ -32,7 +32,7 @@ class HasConstraints(HasVariables[T]):
             properties `h_lbx` and `h_ubx` are called. See these two properties for more
             details. By default, `True`.
         """
-        super().__init__(sym_type=sym_type)
+        super().__init__(sym_type)
 
         self._dual_vars: Dict[str, T] = {}
         self._pars: Dict[str, T] = {}
@@ -42,7 +42,6 @@ class HasConstraints(HasVariables[T]):
         self._lam_lbx, self._lam_ubx = self._csXX(), self._csXX()
         self._g, self._lam_g = self._csXX(), self._csXX()
         self._h, self._lam_h = self._csXX(), self._csXX()
-        self._lbg, self._lbh = np.asarray([]), np.asarray([])
 
         self._remove_redundant_x_bounds = remove_redundant_x_bounds
 
@@ -142,15 +141,15 @@ class HasConstraints(HasVariables[T]):
         """Gets the dual variables of the NLP scheme in vector form.
 
         Note: The order of the dual variables can be adjusted via the class attribute
-        `_DUAL_VARIABLES_ORDER`."""
+        `dual_variables_order`."""
         items = {
             "g": self._lam_g,
             "h": self._lam_h,
             "h_lbx": self.h_lbx[1],
             "h_ubx": self.h_ubx[1],
         }
-        dual = cs.vertcat(*(items.pop(v) for v in self._DUAL_VARIABLES_ORDER))
-        assert not items, "Internal error. _DUAL_VARIABLES_ORDER modified."
+        dual = cs.vertcat(*(items.pop(v) for v in self.dual_variables_order))
+        assert not items, "Internal error. `dual_variables_order` modified."
         return dual
 
     @cached_property
@@ -160,15 +159,15 @@ class HasConstraints(HasVariables[T]):
         property is equivalent to the `lam` property.
 
         Note: The order of the dual variables can be adjusted via the class attribute
-        `_DUAL_VARIABLES_ORDER`."""
+        `dual_variables_order`."""
         items = {
             "g": self._lam_g,
             "h": self._lam_h,
             "h_lbx": self._lam_lbx,
             "h_ubx": self._lam_ubx,
         }
-        dual = cs.vertcat(*(items.pop(v) for v in self._DUAL_VARIABLES_ORDER))
-        assert not items, "Internal error. _DUAL_VARIABLES_ORDER modified."
+        dual = cs.vertcat(*(items.pop(v) for v in self.dual_variables_order))
+        assert not items, "Internal error. `dual_variables_order` modified."
         return dual
 
     def primal_dual_vars(self, all: bool = False) -> T:
@@ -194,7 +193,7 @@ class HasConstraints(HasVariables[T]):
         Note
         ----
         The order of the dual variables can be adjusted via the class attribute
-        `_DUAL_VARIABLES_ORDER`.
+        `dual_variables_order`.
         """
         return cs.vertcat(self._x, self.lam_all if all else self.lam)
 
@@ -317,18 +316,15 @@ class HasConstraints(HasVariables[T]):
         shape = expr.shape
         if op == "==":
             is_eq = True
-            lb = np.zeros(np.prod(shape))
             if soft:
                 raise NotImplementedError(
                     "Soft equality constraints are not currently supported."
                 )
         elif op == "<=":
             is_eq = False
-            lb = np.full(np.prod(shape), -np.inf)
         elif op == ">=":
             is_eq = False
             expr = -expr
-            lb = np.full(np.prod(shape), -np.inf)
         else:
             raise ValueError(f"Unrecognized operator {op}.")
 
@@ -337,14 +333,11 @@ class HasConstraints(HasVariables[T]):
             expr -= slack
 
         self._cons[name] = expr
-        group, con, lam = (
-            ("_g", "_lbg", "_lam_g") if is_eq else ("_h", "_lbh", "_lam_h")
-        )
+        group, lam = ("_g", "_lam_g") if is_eq else ("_h", "_lam_h")
         name_lam = f"{lam[1:]}_{name}"
         lam_c = self._csXX.sym(name_lam, *shape)
         self._dual_vars[name_lam] = lam_c
 
         setattr(self, group, cs.vertcat(getattr(self, group), cs.vec(expr)))
-        setattr(self, con, np.concatenate((getattr(self, con), lb)))
         setattr(self, lam, cs.vertcat(getattr(self, lam), cs.vec(lam_c)))
         return (expr, lam_c, slack) if soft else (expr, lam_c)
