@@ -1,4 +1,4 @@
-from typing import Dict, Literal, Optional, Tuple, TypeVar, Union
+from typing import Callable, Dict, Literal, Optional, Tuple, TypeVar, Union
 
 import casadi as cs
 import numpy as np
@@ -26,13 +26,25 @@ class NlpSensitivity(Wrapper[T]):
         Springer, Berlin, Heidelberg.
     """
 
-    def __init__(self, nlp: Nlp[T], include_barrier_term: bool = True) -> None:
+    def __init__(
+        self,
+        nlp: Nlp[T],
+        target_parameters: Optional[T] = None,
+        include_barrier_term: bool = True,
+    ) -> None:
         """Instantiates the wrapper for performing NLP sensitivities.
 
         Parameters
         ----------
         nlp : Nlp
             The NLP problem to be wrapped.
+        target_parameters : casadi.SX or MX, optional
+            If provided, computes the sensitibity only w.r.t. these parameters. Of
+            course, these parameters must be a subset of all the NLP's parameters. If
+            `None`, then the derivatives and sensitivity are computed w.r.t. all
+            parameters.
+            If new parameters are added after wrapping the nlp in this, be sure to call
+            `set_target_parameters` again.
         include_barrier_term : bool, optional
             If `True`, includes in the KKT matrix a new symbolic variable that
             represents the barrier function of the interior-point solver. Otherwise, no
@@ -41,17 +53,24 @@ class NlpSensitivity(Wrapper[T]):
         """
         super().__init__(nlp)
         self.include_barrier_term = include_barrier_term
+        self.set_target_parameters(target_parameters)
         self._tau = nlp.sym_type.sym("tau") if include_barrier_term else 0
 
+    @property
+    def target_parameters(self) -> T:
+        """Gets the parameters of the NLP that are the target of the sensitivity
+        wrapper, i.e., derivatives and sensitivities are limited to these parameters."""
+        return self._p_idx[0]
+
     @cached_property
     def lagrangian(self) -> T:
         """Gets the Lagrangian of the NLP problem (usually, `L`)."""
         h_lbx, lam_h_lbx = self.nlp.h_lbx
         h_ubx, lam_h_ubx = self.nlp.h_ubx
         return (
-            self.nlp._f
-            + cs.dot(self.nlp._lam_g, self.nlp._g)
-            + cs.dot(self.nlp._lam_h, self.nlp._h)
+            self.nlp.f
+            + cs.dot(self.nlp.lam_g, self.nlp.g)
+            + cs.dot(self.nlp.lam_h, self.nlp.h)
             + cs.dot(lam_h_lbx, h_lbx)
             + cs.dot(lam_h_ubx, h_ubx)
         )
@@ -80,36 +99,41 @@ class NlpSensitivity(Wrapper[T]):
         `_PRIMAL_DUAL_ORDER`.
         """
         items = {
-            "g": self.nlp._g,
-            "h": (self.nlp._lam_h * self.nlp._h) + self._tau,
+            "g": self.nlp.g,
+            "h": (self.nlp.lam_h * self.nlp.h) + self._tau,
             "h_lbx": (self.nlp.h_lbx[0] * self.nlp.h_lbx[1]) + self._tau,
             "h_ubx": (self.nlp.h_ubx[0] * self.nlp.h_ubx[1]) + self._tau,
         }
         kkt = cs.vertcat(
-            cs.jacobian(self.lagrangian, self.nlp._x).T,
-            *(items.pop(v) for v in self.nlp._DUAL_VARIABLES_ORDER)
+            cs.jacobian(self.lagrangian, self.nlp.x).T,
+            *(items.pop(v) for v in self.nlp.dual_variables_order)
         )
-        assert not items, "Internal error. _DUAL_VARIABLES_ORDER modified."
+        assert not items, "Internal error. `dual_variables_order` modified."
         return kkt, (self._tau if self.include_barrier_term else None)
 
     @cached_property
     def jacobians(self) -> Dict[str, T]:
         """Computes various partial derivatives, which are then grouped in a dict with
         the following entries
+            - `L-x`: lagrangian w.r.t. primal variables
             - `L-p`: lagrangian w.r.t. parameters
             - `g-x`: equality constraints w.r.t. primal variables
             - `h-x`: inequality constraints w.r.t. primal variables
             - `K-p`: kkt conditions w.r.t. parameters
             - `K-y`: kkt conditions w.r.t. primal-dual variables
         """
+        L = self.lagrangian
         K = self.kkt[0]
-        y, idx = self._y_idx
+        x = self.nlp.x
+        y, y_idx = self._y_idx
+        p, p_idx = self._p_idx
         return {
-            "L-p": cs.jacobian(self.lagrangian, self.nlp._p),
-            "g-x": cs.jacobian(self.nlp._g, self.nlp._x),
-            "h-x": cs.jacobian(self.nlp._h, self.nlp._x),
-            "K-p": cs.jacobian(K, self.nlp._p),
-            "K-y": cs.jacobian(K, y)[:, idx],
+            "L-x": cs.jacobian(L, x).T,
+            "L-p": cs.jacobian(L, p)[:, p_idx].T,
+            "g-x": cs.jacobian(self.nlp.g, x),
+            "h-x": cs.jacobian(self.nlp.h, x),
+            "K-p": cs.jacobian(K, p)[:, p_idx],
+            "K-y": cs.jacobian(K, y)[:, y_idx],
         }
 
     @cached_property
@@ -120,13 +144,14 @@ class NlpSensitivity(Wrapper[T]):
             - `L-xx`: lagrangian w.r.t. primal variables (twice)
             - `L-px`: lagrangian w.r.t. parameters and then primal variables
         """
-        L = self.lagrangian
-        Lpp, Lp = cs.hessian(L, self.nlp._p)
-        Lpx = cs.jacobian(Lp, self.nlp._x)
+        x = self.nlp.x
+        p, p_idx = self._p_idx
+        Lx = self.jacobians["L-x"]
+        Lp = self.jacobians["L-p"]
         return {
-            "L-pp": Lpp,
-            "L-xx": cs.hessian(L, self.nlp._x)[0],
-            "L-px": Lpx,
+            "L-xx": cs.jacobian(Lx, x),
+            "L-pp": cs.jacobian(Lp, p)[:, p_idx],
+            "L-px": cs.jacobian(Lp, x),
         }
 
     @cached_property
@@ -141,12 +166,13 @@ class NlpSensitivity(Wrapper[T]):
         jacobians = self.jacobians
         Kp = jacobians["K-p"]
         Ky = jacobians["K-y"]
-        y, idx = self._y_idx
+        y, y_idx = self._y_idx
+        p, p_idx = self._p_idx
         return {
-            "K-pp": hojacobian(Kp, self.nlp._p)[..., 0],
-            "K-yp": hojacobian(Ky, self.nlp._p)[..., 0],
-            "K-yy": hojacobian(Ky, y)[..., idx, 0],
-            "K-py": hojacobian(Kp, y)[..., idx, 0],
+            "K-pp": hojacobian(Kp, p)[..., p_idx, 0],
+            "K-yp": hojacobian(Ky, p)[..., p_idx, 0],
+            "K-yy": hojacobian(Ky, y)[..., y_idx, 0],
+            "K-py": hojacobian(Kp, y)[..., y_idx, 0],
         }
 
     @property
@@ -171,18 +197,19 @@ class NlpSensitivity(Wrapper[T]):
     def parametric_sensitivity(
         self,
         expr: T = None,
-        solution: Optional[Solution] = None,
+        solution: Optional[Solution[T]] = None,
+        second_order: bool = False,
     ) -> Union[
-        Tuple[T, T],
-        Tuple[cs.DM, cs.DM],
-        Tuple[npt.NDArray[np.double], npt.NDArray[np.double]],
+        Tuple[T, Optional[T]],
+        Tuple[cs.DM, Optional[cs.DM]],
+        Tuple[npt.NDArray[np.double], Optional[npt.NDArray[np.double]]],
     ]:
         """Performs the (symbolic or numerical) sensitivity of the NLP w.r.t. its
         parametrization, according to [1].
 
         Parameters
         ----------
-        expression : casadi.SX or MX, optional
+        expr : casadi.SX or MX, optional
             If provided, computes the sensitivity of this expression (which must be
             dependent on the primal-dual variables and/or parameters of the NLP) w.r.t.
             the NLP parameters. If `None`, then the sensitivity of the primal-dual
@@ -191,12 +218,16 @@ class NlpSensitivity(Wrapper[T]):
             If a solution is passed, then the sensitivity is numerically computed for
             that solution; otherwise, the sensitivity is carried out symbolically
             (however, this is much more computationally intensive).
+        second_order : bool, optional
+            If `second_order=False`, the analysis is stopped at the first order;
+            otherwise, also second-order information is computed.
 
         Returns
         -------
         2-element tuple of casadi.SX, or MX, or DM, or arrays
             The 1st and 2nd-order NLP parametric sensitivity in the form of an array/DM,
-            if a solution is passed, or a symbolic vector SX/MX.
+            if a solution is passed, or a symbolic vector SX/MX. When
+            `second_order=False` the second element in the tuple is `None`.
 
         Raises
         ------
@@ -210,71 +241,101 @@ class NlpSensitivity(Wrapper[T]):
             S.O. Krumke, and J. Rambau (eds.), Online Optimization of Large Scale
             Systems, 3–16. Springer, Berlin, Heidelberg.
         """
+        # first and second order sensitivities, a.k.a., dydp and d2dydp2
+        d: Callable[[T], Union[T, cs.DM]] = (
+            (lambda o: o)
+            if solution is None
+            else (lambda o: solution.value(o))  # type: ignore
+        )
+        dydp, dydp_np, d2ydp2 = self._y_parametric_sensitivity(
+            solution, second_order, d
+        )
+
+        # first order sensitivity of Z, a.k.a., dZdp
+        Z = expr  # Z := z(x(p),lam(p),p)
+        if Z is not None:
+            Zshape = Z.shape
+            Z = cs.vec(Z)
+            y, y_idx = self._y_idx
+            p, p_idx = self._p_idx
+            np_ = p.shape[0] if isinstance(p_idx, slice) else p_idx.size
+
+            if second_order:
+                Zpp, Zp = hohessian(Z, p)
+                Zyy, Zy = hohessian(Z, y)
+            else:
+                Zp = hojacobian(Z, p)
+                Zy = hojacobian(Z, y)
+            Zp = d(array2cs(Zp[:, 0, p_idx, 0]))
+            Zy = d(array2cs(Zy[:, 0, y_idx, 0]))
+
+            dZdp = Zy @ dydp + Zp
+            dZdp = cs2array(dZdp).reshape(Zshape + (np_,), order="F").squeeze()
+            if dZdp.ndim <= 2:
+                dZdp = array2cs(dZdp)
+
+        # if no second order is required, just return
+        if not second_order:
+            return dydp if Z is None else dZdp, None
+
+        # if no expression is given, just return
+        if Z is None:
+            d2ydp2 = d2ydp2.squeeze()
+            if d2ydp2.ndim <= 2:
+                d2ydp2 = array2cs(d2ydp2)
+            return dydp, d2ydp2
+
+        # second order sensitivity of Z, a.k.a., d2Zdp2
+        Zyp = d(hohessian(Z, y, p)[0][:, 0, y_idx, 0][:, :, p_idx, 0])
+        Zpy = d(hohessian(Z, p, y)[0][:, 0, p_idx, 0][:, :, y_idx, 0])
+        Zpp = d(Zpp[:, 0, p_idx, 0][:, :, p_idx, 0])
+        Zyy = d(Zyy[:, 0, y_idx, 0][:, :, y_idx, 0])
+        T1 = (d2ydp2.transpose((1, 2, 0)) @ cs2array(Zy).T).transpose((2, 0, 1))
+        T2 = Zyy.transpose((0, 2, 1)) @ dydp_np + Zpy.transpose((0, 2, 1)) + Zyp
+        d2Zdp2 = Zpp + T1 + dydp_np.T @ T2
+        d2Zdp2 = d2Zdp2.reshape(Zshape + (np_, np_), order="F").squeeze()
+        if d2Zdp2.ndim <= 2:
+            d2Zdp2 = array2cs(d2Zdp2)
+        return dZdp, d2Zdp2
+
+    def _y_parametric_sensitivity(
+        self,
+        solution: Solution,
+        second_order: bool,
+        d: Callable[[T], Union[T, cs.DM]],
+    ) -> Union[
+        Tuple[T, np.ndarray, T],
+        Tuple[cs.DM, np.ndarray, cs.DM],
+        Tuple[npt.NDArray[np.double], np.ndarray, npt.NDArray[np.double]],
+    ]:
+        """Internal utility to compute the sensitivity of y w.r.t. p."""
         # first order sensitivity, a.k.a., dydp
-        Ky = self.jacobians["K-y"]
-        Kp = self.jacobians["K-p"]
+        Ky = d(self.jacobians["K-y"])
+        Kp = d(self.jacobians["K-p"])
         if solution is None:
-            dydp = -cs.inv(Ky) @ Kp
+            Ky_inv = cs.inv(Ky)
+            dydp = -Ky_inv @ Kp
         else:
-            A = solution.value(Ky)
-            b = -solution.value(Kp)
-            dydp = np.linalg.solve(A, b)
+            dydp = -np.linalg.solve(Ky, Kp)
+        if not second_order:
+            return dydp, None, None  # type: ignore
 
         # second order sensitivity, a.k.a., d2ydp2
-        dydp = cs2array(dydp)
-        Kpp = self.hojacobians["K-pp"]
-        Kpy = self.hojacobians["K-py"]
-        Kyp = self.hojacobians["K-yp"]
-        Kyy = self.hojacobians["K-yy"]
+        dydp_ = cs2array(dydp)
+        Kpp = d(self.hojacobians["K-pp"])
+        Kpy = d(self.hojacobians["K-py"])
+        Kyp = d(self.hojacobians["K-yp"])
+        Kyy = d(self.hojacobians["K-yy"])
         M = (
             Kpp
-            + (Kpy.transpose((0, 2, 1)) + Kyp + (Kyy @ dydp)).transpose((0, 2, 1))
-            @ dydp
-        )
+            + (Kpy.transpose((0, 2, 1)) + Kyp + (Kyy @ dydp_)).transpose((0, 2, 1))
+            @ dydp_
+        ).transpose((2, 0, 1))
         if solution is None:
-            A = cs2array(cs.inv(Ky))
-            b = -M.transpose((2, 0, 1))
-            d2ydp2 = (A @ b).transpose((1, 2, 0))
+            d2ydp2 = -(cs2array(cs.inv(Ky)) @ M).transpose((1, 2, 0))
         else:
-            A = solution.value(Ky)
-            b = -solution.value(M).transpose((2, 0, 1))
-            d2ydp2 = np.linalg.solve(A, b).transpose((1, 2, 0))
-
-        # sensitivity of custom expression
-        Z = expr  # Z := z(x(p),lam(p),p)
-        if Z is None:
-            dydp = dydp.squeeze()
-            d2ydp2 = d2ydp2.squeeze()
-            if solution is not None:
-                return dydp, d2ydp2
-            return (array2cs(dydp), array2cs(d2ydp2) if d2ydp2.ndim <= 2 else d2ydp2)
-        Zshape = Z.shape
-        Z = cs.vec(Z)
-
-        p = self.nlp._p
-        y, idx = self._y_idx
-        Zpp, Zp = hohessian(Z, p)
-        Zyy, Zy = hohessian(Z, y)
-        Zyp = hohessian(Z, y, p)[0]
-        Zpy = hohessian(Z, p, y)[0]
-        Zpp = Zpp.squeeze((1, 3, 5))
-        Zp = Zp.squeeze((1, 3))
-        Zy = Zy[:, 0, idx, 0]
-        Zyy = Zyy.squeeze((1, 3, 5))[:, idx, :][..., idx]
-        Zyp = Zyp.squeeze((1, 3, 5))[:, idx, :]
-        Zpy = Zpy.squeeze((1, 3, 5))[..., idx]
-
-        dzdp = Zy @ dydp + Zp
-        T1 = (d2ydp2.transpose((1, 2, 0)) @ Zy.T).transpose((2, 0, 1))
-        T2 = Zyy.transpose((0, 2, 1)) @ dydp + Zpy.transpose((0, 2, 1)) + Zyp
-        d2zdp2 = Zpp + T1 + dydp.T @ T2
-
-        np_ = self.nlp.np
-        dzdp = dzdp.reshape(Zshape + (np_,), order="F")
-        d2zdp2 = d2zdp2.reshape(Zshape + (np_, np_), order="F")
-        if solution is not None:
-            return solution.value(dzdp), solution.value(d2zdp2)
-        return dzdp, d2zdp2
+            d2ydp2 = -np.linalg.solve(Ky, M).transpose((1, 2, 0))
+        return dydp, dydp_, d2ydp2
 
     @invalidate_cache(jacobians, hessians, hojacobians)
     def parameter(self, name: str, shape: Tuple[int, int] = (1, 1)) -> T:
@@ -310,22 +371,56 @@ class NlpSensitivity(Wrapper[T]):
         """See `Nlp.minimize` method."""
         return self.nlp.minimize(objective)
 
+    @invalidate_cache(jacobians, hessians, hojacobians)
+    def set_target_parameters(self, parameters: Optional[T]) -> None:
+        """Sets the target parameters of the sensitivity wrapper.
+
+        Parameters
+        ----------
+        parameters : casadi.MX or SX or None
+            New parameters to target during the sensitivity analyses. If `None`, all
+            NLP parameters are included.
+        """
+        if parameters is None:
+            self._p_idx_internal = None, slice(None)
+            return
+
+        p: T = cs.vec(parameters)
+        if self.nlp.sym_type is cs.SX:
+            self._p_idx_internal = p, slice(None)
+            return
+
+        sp_J: cs.Sparsity = cs.jacobian(p, self.nlp.p).sparsity()
+        idx = np.asarray(sp_J.get_crs()[1], dtype=int)
+        assert idx.size == p.shape[0], (
+            "Invalid subset of target parameters (some were not found in the"
+            " original NLP parameters)."
+        )
+        self._p_idx_internal = self.nlp.p, idx  # type: ignore
+
     @property
     def _y_idx(self) -> Tuple[T, Union[slice, npt.NDArray[np.int64]]]:
         """Internal utility to return all the primal-dual variables and indices that are
         associated to non-redundant entries in the kkt conditions."""
-        if self.nlp._csXX is cs.SX:
-            y = self.nlp.primal_dual_vars()
-            idx = slice(None)
-        else:
-            # in case of MX, jacobians throw if the MX are indexed (no more
-            # symbolical according to the exception). So we take the jacobian
-            # with all primal-dual vars, and then index the relevant rows/cols.
-            y = self.nlp.primal_dual_vars(all=True)
-            h_lbx_idx = np.where(self.nlp._lbx != -np.inf)[0]
-            h_ubx_idx = np.where(self.nlp._ubx != +np.inf)[0]
-            n = self.nlp.nx + self.nlp.ng + self.nlp.nh
-            idx = np.concatenate(
-                (np.arange(n), h_lbx_idx + n, h_ubx_idx + n + h_lbx_idx.size)
-            )
+        if self.nlp.sym_type is cs.SX:
+            return self.nlp.primal_dual_vars(), slice(None)
+
+        # in case of MX, jacobians throw if the MX are indexed (no more
+        # symbolical according to the exception). So we take the jacobian
+        # with all primal-dual vars, and then index the relevant rows/cols.
+        y = self.nlp.primal_dual_vars(all=True)
+        h_lbx_idx = np.where(self.nlp.lbx != -np.inf)[0]
+        h_ubx_idx = np.where(self.nlp.ubx != +np.inf)[0]
+        n = self.nlp.nx + self.nlp.ng + self.nlp.nh
+        idx = np.concatenate(
+            (np.arange(n), h_lbx_idx + n, h_ubx_idx + n + h_lbx_idx.size)
+        )
         return y, idx
+
+    @property
+    def _p_idx(self) -> Tuple[T, Union[slice, npt.NDArray[np.int64]]]:
+        """Internal utility to return the indices of p from all the NLP pars. Like for
+        _y_idx, SX is fine with jacobians, but MX jacobians need to be computed for
+        all elements and then indexed."""
+        p = self._p_idx_internal[0]
+        return (self.nlp.p if p is None else p), self._p_idx_internal[1]
